class PolicyContext:
    """
    A class containing data about the context of an applied authorization policy.

    Authorization policies use this class to influence access determination.

    Properties
    ----------
    access_verdict : bool | None
        The verdict that will influence access determination.
    """

    def __init__(self):
        self._access_verdict = None

    @property
    def access_verdict(self):
        """The verdict that will influence access determination.

        By default None.

        If True: access will be allowed.
        If False: access will be denied.
        If None: access will be denied if no other policy allows.
        """

        return self._access_verdict

    @access_verdict.setter
    def access_verdict(self, access_verdict: bool | None):
        self._access_verdict = access_verdict



import os
import requests

from . import flask_process, custom_token, helpers
from dotenv import load_dotenv


load_dotenv()


class TestPolicyFactory:
    ENDPOINT = str(os.getenv("ENDPOINT"))
    SUPER_ADMIN_ROLE = "super_admin"
    REGULAR_USER_ROLE = "regular_user"

    @classmethod
    def setup_class(cls):
        flask_process.set_app_policies(
            [
                "token_based_policies.authlib_flask_oauth2_policy",
                "token_based_policies.tenant_token_roles_policy",
            ],
            ["super_admin_policy", "open_data_policy", "scope_based_policy"],
        )
        flask_process.start()

    def setup_method(self):
        flask_process.assert_running()

    def test_single_request_without_token_returns_401(self):
        response = requests.put(self.ENDPOINT)
        assert response.status_code == 401

    def test_request_with_wrong_token_payload_structure_returns_403(self):
        payload = {
            "azp": "inuits-policy-based-auth",
            "inuits-policy-based-auth": {"roles": [self.SUPER_ADMIN_ROLE]},
        }
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_with_wrong_token_azp_returns_403(self):
        payload = {
            "azp": "inuits-policy-based-auth",
            "resource_access": {
                "policy-based-auth": {"roles": [self.SUPER_ADMIN_ROLE]}
            },
        }
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_without_token_returns_401_after_a_successful_request_with_token(
        self,
    ):
        payload = helpers.get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)
        assert response.status_code == 200

        response = requests.put(self.ENDPOINT)
        assert response.status_code == 401

    def test_user_context_can_be_modified_in_authorization_policies(self):
        payload = helpers.get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.put(self.ENDPOINT, headers=headers)
        response_body = response.json()
        assert response.status_code == 200
        assert response_body["bag"] == {}
        assert response_body["access_restrictions"]["filters"] is None

        response = requests.get(self.ENDPOINT, headers=headers)
        response_body = response.json()
        assert response.status_code == 200
        assert response_body["bag"] == {"test": "test"}
        assert response_body["access_restrictions"]["filters"] == {
            "relations.hasTenant": ""
        }

    def test_user_context_is_cleared_correctly(self):
        payload = helpers.get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)
        response_body = response.json()
        assert response.status_code == 200
        assert response_body["bag"] == {"test": "test"}
        assert response_body["access_restrictions"]["filters"] == {
            "relations.hasTenant": ""
        }

        response = requests.put(self.ENDPOINT, headers=headers)
        response_body = response.json()
        assert response.status_code == 200
        assert response_body["bag"] == {}
        assert response_body["access_restrictions"]["filters"] is None

    @classmethod
    def teardown_class(cls):
        flask_process.stop()



from abc import ABC, abstractmethod
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.exceptions import (
    AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException,
)


class BaseAuthorizationPolicy(ABC):
    """
    An abstract class used as an interface for concrete implementations of
    authorization policies.

    Methods
    -------
    apply(policy_context, user_context, request_context):
        Applies the policy.

    authorize(policy_context, user_context, request_context):
        Authorizes a user.
    """

    def apply(
        self,
        policy_context: PolicyContext,
        user_context: UserContext,
        request_context: RequestContext,
    ) -> PolicyContext:
        """Applies the policy.

        Sets policy_context.access_verdict to its default value None before authorizing.

        Parameters
        ----------
        policy_context : PolicyContext
            An object containing data about the context of an applied
            authorization policy.
        user_context : UserContext
            An object containing data about the authenticated user.
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        PolicyContext
            An object containing data about the context of an applied
            authorization policy.

        Raises
        ------
        AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException:
            If the authorize method does not return an object of type
            PolicyContext.
        """

        policy_context.access_verdict = None
        policy_context = self.authorize(policy_context, user_context, request_context)
        if not isinstance(policy_context, PolicyContext):
            raise AuthorizeMethodDidNotReturnObjectOfTypePolicyContextException()

        return policy_context

    @abstractmethod
    def authorize(
        self,
        policy_context: PolicyContext,
        user_context: UserContext,
        request_context: RequestContext,
    ) -> PolicyContext:
        """Authorizes a user.

        Parameters
        ----------
        policy_context : PolicyContext
            An object containing data about the context of an applied
            authorization policy.
        user_context : UserContext
            An object containing data about the authenticated user.
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        PolicyContext
            An object containing data about the context of an applied
            authorization policy.
        """

        pass



import base64
import json
import requests

from abc import ABC
from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.jose import jwt
from authlib.oauth2 import OAuth2Error
from authlib.oauth2.rfc6750 import BearerTokenValidator, InvalidTokenError
from authlib.oauth2.rfc7523 import JWTBearerToken
from inuits_policy_based_auth.authentication.base_authentication_policy import (
    BaseAuthenticationPolicy,
)
from logging import Logger
from werkzeug.exceptions import Unauthorized


class AuthlibFlaskOauth2Policy(BaseAuthenticationPolicy):
    """
    An authentication policy that uses Authlib Flask OAuth2 to do token-based
    authentication.

    Parameters:
    -----------
    logger : Logger
        Logger object for logging authentication events and errors.
    token_schema : dict
        Dict containing mappings between property <-> path.to.that.property.in.token.
    static_issuer : str, optional
        A string representing the issuer of the JWT. This parameter is required
        if remote token validation is not enabled.
    static_public_key : str, optional
        A string representing the public key used to verify the signature of the
        JWT. This parameter is required if remote token validation is not enabled.
    allowed_issuers : List[str], optional
        A list of token issuers whose tokens are allowed. If this parameter is not
        passed or the list is empty, all issuers are allowed.
    allow_anonymous_users : bool, optional
        A bool about whether anonymous users are allowed to do requests.
    **kwargs : dict
        Any additional keyword arguments to be passed to the JWTValidator constructor.
    """

    def __init__(
        self,
        logger: Logger,
        token_schema: dict,
        static_issuer=None,
        static_public_key=None,
        allowed_issuers=None,
        allow_anonymous_users=False,
        **kwargs,
    ):
        validator = JWTValidator(
            logger,
            static_issuer,
            static_public_key,
            allowed_issuers,
            **kwargs,
        )
        resource_protector = ResourceProtector()
        resource_protector.register_token_validator(validator)
        self._resource_protector = resource_protector

        self._logger = logger
        self._token_schema = token_schema
        self._allow_anonymous_users = allow_anonymous_users

    def authenticate(self, user_context, _):
        """
        Authenticate the user based on the token in the user context.

        Parameters:
        -----------
        user_context : UserContext
            The context of the user requesting authentication.
        request_context : RequestContext
            The context of the request.

        Returns:
        --------
        UserContext
            The user context with the authenticated user details added.

        Raises:
        -------
        Unauthorized
            If the authentication fails.
        """

        try:
            token = self._resource_protector.acquire_token()
            user_context.auth_objects.add_key_value_pair("token", token)
            flattened_token = user_context.flatten_auth_object(token)

            user_context.id = flattened_token[self._token_schema["id"]]
            user_context.email = flattened_token.get(
                self._token_schema.get("email", ""), ""
            ).lower()
            user_context.preferred_username = flattened_token.get(
                self._token_schema.get("preferred_username", ""), ""
            ).lower()
            return user_context
        except (InvalidTokenError, ValidatorException) as error:
            raise Unauthorized(str(error))
        except OAuth2Error as error:
            if self._allow_anonymous_users:
                return user_context
            else:
                raise Unauthorized(str(error))


class JWTValidator(BearerTokenValidator, ABC):
    """
    Validator for JSON Web Tokens (JWT).

    This class inherits from BearerTokenValidator and validates tokens of type "bearer".
    It also has the ability to perform remote token validation by checking against an
    OpenID Connect provider.

    Parameters:
    -----------
    logger : Logger
        An instance of a logger object to be used for logging.
    static_issuer : str, optional
        A string representing the issuer of the JWT. This parameter is required
        if remote token validation is not enabled.
    static_public_key : str, optional
        A string representing the public key used to verify the signature of the
        JWT. This parameter is required if remote token validation is not enabled.
    allowed_issuers : List[str], optional
        A list of token issuers whose tokens are allowed. If this parameter is not
        passed or the list is empty, all issuers are allowed.
    **kwargs : dict
        Additional keyword arguments to be passed to the parent class.

    Attributes:
    -----------
    TOKEN_TYPE : str
        A string representing the token type that this validator is responsible for.
        Always "bearer".
    token_cls : type
        A class representing the type of token that this validator is responsible for.
        Always JWTBearerToken.
    claims_options : dict
        A dictionary representing the claims options to be used for validating the JWT.
        By default, this dictionary contains options for the "exp", "azp", and "sub" claims.

    Methods:
    --------
    authenticate_token(token_string: str) -> Optional[JWTBearerToken]:
        Authenticates a JWT token and returns a JWTBearerToken object if successful. If
        authentication fails, None is returned.

    Private Methods:
    ----------------
    __get_issuer_from_token_string(token_string: str) -> Optional[str]:
        Extracts the issuer from a JWT token without verifying the signature.

    __get_jwks_from_issuer(issuer: str) -> dict:
        Retrieves JWKS from the issuer to be used to validate the token.
    """

    TOKEN_TYPE = "bearer"
    token_cls = JWTBearerToken

    def __init__(
        self,
        logger,
        static_issuer,
        static_public_key,
        allowed_issuers,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.logger = logger
        self.static_issuer = static_issuer
        self.static_public_key = static_public_key
        self.allowed_issuers = allowed_issuers
        self.claims_options = {
            "exp": {"essential": True},
            "azp": {"essential": True},
            "sub": {"essential": True},
        }
        self.jwks_cache = {}

    def __decode_token(self, token_string, jwks):
        try:
            claims = jwt.decode(
                token_string,
                jwks,
                claims_options=self.claims_options,
                claims_cls=self.token_cls,
            )
            claims.validate()
            return claims
        except:
            return None

    def __get_jwks(self, issuer):
        if issuer == self.static_issuer:
            jwks = f"-----BEGIN PUBLIC KEY-----\n{self.static_public_key}\n-----END PUBLIC KEY-----"
        elif issuer in self.jwks_cache and self.jwks_cache[issuer] is not None:
            jwks = self.jwks_cache[issuer]
        else:
            jwks = self.__get_jwks_from_issuer(issuer)
            self.jwks_cache[issuer] = jwks
        return jwks

    def authenticate_token(self, token_string):
        """
        Authenticate a JWT token and return a JWTBearerToken object if successful.

        Parameters:
        -----------
        token_string : str
            A string representing the JWT token to be authenticated.

        Returns:
        --------
        Optional[JWTBearerToken]
            If authentication is successful, returns a JWTBearerToken object containing
            the decoded JWT claims. If authentication fails, None is returned.
        """
        issuer = self.__get_issuer_from_token_string(token_string)
        if not self.allowed_issuers:
            raise ValidatorException(f"No allowed issuers configured")
        elif issuer not in self.allowed_issuers:
            raise ValidatorException(f"Issuer {issuer} not allowed")
        jwks = self.__get_jwks(issuer)
        token = self.__decode_token(token_string, jwks)
        if not token:
            self.jwks_cache[issuer] = None
            jwks = self.__get_jwks(issuer)
            token = self.__decode_token(token_string, jwks)
        return token

    @staticmethod
    def __get_issuer_from_token_string(token_string):
        # Adding "=="  is necessary for correct base64 padding
        payload = f'{token_string.split(".")[1]}=='
        decoded = json.loads(base64.urlsafe_b64decode(payload.encode("utf-8")))
        return decoded.get("iss")

    @staticmethod
    def __get_jwks_from_issuer(issuer):
        req = requests.get(f"{issuer}/.well-known/openid-configuration")
        if req.status_code != 200:
            raise ValidatorException(
                f"Failed to get issuer's OpenID configuration: {req.text.strip()}"
            )
        jwks_url = req.json()["jwks_uri"]
        req = requests.get(jwks_url)
        if req.status_code != 200:
            raise ValidatorException(f"Failed to get issuer's JWKS: {req.text.strip()}")
        return req.json()["keys"]


class ValidatorException(Exception):
    def __init__(self, message):
        super().__init__(message)



class RequestContext:
    """A class containing data about the context of a request.

    The data provided in this class is used in policies to determine whether a
    user is authenticated/authorized.

    Properties
    ----------
    http_request : Unknown
        An HTTP request used by policies to determine access to a resource.
    resource_scopes : list[str], optional
        Scopes a user must have for accessing a protected resource when policies
        rely on scope-based authorization.
    """

    def __init__(self, http_request, resource_scopes: list[str] = None):
        """
        Parameters
        ----------
        http_request : Unknown
            An HTTP request used by policies to determine access to a resource.
        resource_scopes : list[str], optional
            Scopes a user must have for accessing a protected resource when policies
            rely on scope-based authorization.
        """

        self._http_request = http_request
        self._resource_scopes = resource_scopes if resource_scopes else []

    @property
    def http_request(self):
        """
        An HTTP request used by policies to determine access to a resource.
        """

        return self._http_request

    @property
    def resource_scopes(self):
        """
        Scopes a user must have for accessing a protected resource when policies
        rely on scope-based authorization.
        """

        return self._resource_scopes

    def _serialize(self) -> dict:
        return {
            "http_request": {
                "method": self._http_request.method,
                "path": self._http_request.path,
                "headers": self._http_request.headers,
                "data": self._http_request.data,
            }
        }

    def __hash__(self) -> int:
        return hash(str(self._serialize()))

    def __eq__(self, request_context) -> bool:
        return self.__hash__() == request_context.__hash__()

    def __ne__(self, request_context) -> bool:
        return not self.__eq__(request_context)



from inuits_policy_based_auth.authorization.policies.super_admin_policy import (
    SuperAdminPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.helpers.tenant import Tenant


class TestSuperAdminPolicy:
    def setup_method(self):
        self.super_admin_policy = SuperAdminPolicy()
        self.policy_context = PolicyContext()
        self.user_context = UserContext()
        self.request_context = RequestContext(None)
        self.user_context.x_tenant = Tenant()

    def test_authorize_allows_access(self):
        self.user_context.x_tenant.roles = ["super_admin"]  # pyright: ignore
        policy_context = self.super_admin_policy.authorize(
            self.policy_context, self.user_context, self.request_context
        )
        assert policy_context.access_verdict == True

    def test_authorize_does_not_determine_access(self):
        self.user_context.x_tenant.roles = ["regular_user"]  # pyright: ignore
        policy_context = self.super_admin_policy.authorize(
            self.policy_context, self.user_context, self.request_context
        )
        assert policy_context.access_verdict is None



import pytest

from inuits_policy_based_auth.helpers.immutable_dict import ImmutableDict


class TestImmutableDict:
    def setup_method(self):
        self.data = {"key1": "value1", "key2": {"inner_key1": "inner_value1"}}

    def test_init(self):
        immutable_dict = ImmutableDict(self.data)
        assert str(immutable_dict._data) == str(self.data)

    def test_init_raises_TypeError_if_not_dict_like(self):
        with pytest.raises(TypeError):
            ImmutableDict("not a dict-like object")

    def test_get_existing_key(self):
        immutable_dict = ImmutableDict(self.data)
        result = immutable_dict.get("key1")
        assert result is not None

    def test_get_non_existing_key(self):
        immutable_dict = ImmutableDict(self.data)
        result = immutable_dict.get("non_existent_key")
        assert result is None

    def test_add_key_value_pair(self):
        immutable_dict = ImmutableDict(self.data)

        immutable_dict.add_key_value_pair("key3", "value3")

        assert str(immutable_dict._data) == str(
            {"key1": "value1", "key2": {"inner_key1": "inner_value1"}, "key3": "value3"}
        )
        assert str(immutable_dict._data) != str(self.data)

    def test_add_key_value_pair_raises_TypeError_if_modified(self):
        immutable_dict = ImmutableDict(self.data)
        with pytest.raises(TypeError):
            immutable_dict.add_key_value_pair("key2", "value2")

    def test_getitem(self):
        immutable_dict = ImmutableDict(self.data)
        assert immutable_dict["key1"] == "value1"
        assert immutable_dict["key2"]["inner_key1"] == "inner_value1"

    def test_getitem_raises_KeyError_if_key_not_found(self):
        immutable_dict = ImmutableDict(self.data)
        with pytest.raises(KeyError):
            immutable_dict["non_existent_key"]

    def test_setitem_raises_TypeError_if_modified(self):
        immutable_dict = ImmutableDict(self.data)
        with pytest.raises(TypeError):
            immutable_dict["key1"] = "new_value"

    def test_delitem_raises_TypeError_if_modified(self):
        immutable_dict = ImmutableDict(self.data)
        with pytest.raises(TypeError):
            del immutable_dict["key"]

    def test_immutable_dict_has_string_representation(self):
        immutable_dict = ImmutableDict(self.data)
        assert repr(immutable_dict) == str(self.data)

    def test_initial_data_is_added_recursively(self):
        data = {
            "key1": "value1",
            "key2": {"key3": "value3", "key4": [1, 2, {"key5": "value5"}]},
        }

        immutable_dict = ImmutableDict(data)

        assert immutable_dict["key1"] == "value1"
        assert immutable_dict["key2"]["key3"] == "value3"
        assert immutable_dict["key2"]["key4"][2]["key5"] == "value5"
        with pytest.raises(TypeError):
            ImmutableDict(1234)

    def test_empty_list(self):
        assert ImmutableDict._freeze_list([]) == tuple()

    def test_frozen_list(self):
        items = [1, 2, 3, {"a": 4, "b": [5, 6]}]

        frozen_items = ImmutableDict._freeze_list(items)

        assert isinstance(frozen_items, tuple)
        assert len(frozen_items) == len(items)
        assert frozen_items[3]["b"][1] == 6
        with pytest.raises(TypeError):
            frozen_items[1]["a"] = 3

    def test_nested_list(self):
        items = [1, [2, [3, [4]]], 5]
        frozen_items = ImmutableDict._freeze_list(items)
        assert isinstance(frozen_items, tuple)
        assert frozen_items[1][1][1][0] == 4



import os

from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)


class SuperAdminPolicy(BaseAuthorizationPolicy):
    """
    An authorization policy that allows a user to do anything if he/she has a super
    admin role.
    """

    super_admin_role: str = os.getenv("SUPER_ADMIN_ROLE", "super_admin")

    def authorize(self, policy_context, user_context, _):
        if self.super_admin_role in user_context.x_tenant.roles:
            policy_context.access_verdict = True

        return policy_context



from collections.abc import MutableMapping
from inuits_policy_based_auth.helpers.access_restrictions import AccessRestrictions
from inuits_policy_based_auth.helpers.immutable_dict import ImmutableDict
from inuits_policy_based_auth.helpers.tenant import Tenant


class UserContext:
    """
    A class containing data about the authenticated user.

    Properties
    ----------
    auth_objects : ImmutableDict
        An immutable dict containing objects used to authenticate a user,
        for example a token.
    id : str
        The id of the authenticated user.
    email : str
        The email of the authenticated user.
    preferred_username : str
        The preferred_username of the authenticated user.
    x_tenant : Tenant
        The user tenant that is requested from the X-Tenant-Id http header.
    tenants : list[Tenant]
        All tenants that are related to the user.
    bag : dict
        Dict that can contain any kind of information. It enables the possibility
        to share values dynamically between policies themselves or between
        policies and the api.
    access_restrictions : AccessRestrictions
        A class containing properties used to restrict access.

    Methods
    -------
    flatten_auth_object(data, parent_key)
        Flattens the auth object to be a dict of one level deep.
    """

    def __init__(self):
        self._auth_objects = ImmutableDict({})
        self._id = ""
        self._email = ""
        self._preferred_username = ""
        self._x_tenant = Tenant()
        self._tenants: list[Tenant] = []
        self._bag = {}
        self._access_restrictions = AccessRestrictions()

    @property
    def auth_objects(self):
        """
        An immutable dict containing objects used to authenticate a user,
        for example a token.
        """

        return self._auth_objects

    @property
    def id(self):
        """The id of the authenticated user."""

        return self._id

    @id.setter
    def id(self, id: str):
        self._id = id

    @property
    def email(self):
        """The email of the authenticated user."""

        return self._email

    @email.setter
    def email(self, email: str):
        self._email = email

    @property
    def preferred_username(self):
        """The preferred_username of the authenticated user."""

        return self._preferred_username

    @preferred_username.setter
    def preferred_username(self, preferred_username: str):
        self._preferred_username = preferred_username

    @property
    def x_tenant(self):
        """The user tenant that is requested from the X-Tenant-Id http header."""

        return self._x_tenant

    @x_tenant.setter
    def x_tenant(self, x_tenant: Tenant):
        self._x_tenant = x_tenant

    @property
    def tenants(self):
        """All tenants that are related to the user."""

        return self._tenants

    @tenants.setter
    def tenants(self, tenants: list[Tenant]):
        self._tenants = tenants

    @property
    def bag(self):
        """
        Dict that can contain any kind of information. It enables the possibility
        to share values dynamically between policies themselves or between
        policies and the api.
        """

        return self._bag

    @bag.setter
    def bag(self, bag: dict):
        self._bag = bag

    @property
    def access_restrictions(self):
        """A class containing properties used to restrict access."""

        return self._access_restrictions

    def __flatten_auth_object_generator(self, data: MutableMapping, parent_key):
        for key, value in data.items():
            flattened_key = parent_key + "." + key if parent_key else key
            if isinstance(value, MutableMapping):
                yield from self.flatten_auth_object(value, flattened_key).items()
            else:
                yield flattened_key, value

    def flatten_auth_object(self, data: MutableMapping, parent_key=""):
        """Flattens the auth object to be a dict of one level deep.

        Parameters
        ----------
        data : MutableMapping
            An object used to authenticate a user with, for example a token.
        parent_key : str, optional
            A key that will be the root key for the flattened dict.

        Returns
        -------
        dict
            A flattened dictionary representation of the auth object.
        """

        return dict(self.__flatten_auth_object_generator(data, parent_key))



import os
import requests

from .. import flask_process, custom_token, helpers
from dotenv import load_dotenv


load_dotenv()


class TestSuperAdminPolicy:
    ENDPOINT = str(os.getenv("ENDPOINT"))
    SUPER_ADMIN_ROLE = "super_admin"

    @classmethod
    def setup_class(cls):
        flask_process.set_app_policies(
            [
                "token_based_policies.authlib_flask_oauth2_policy",
                "token_based_policies.tenant_token_roles_policy",
            ],
            ["super_admin_policy"],
        )
        flask_process.start()

    def setup_method(self):
        flask_process.assert_running()

    def test_token_without_super_admin_role_returns_403(self):
        payload = {"azp": "inuits-policy-based-auth"}
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_with_wrong_role_returns_403(self):
        payload = helpers.get_payload(["regular_user"])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_get_request_with_correct_role_returns_200(self):
        payload = helpers.get_payload([self.SUPER_ADMIN_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        assert response.status_code == 200
        assert response.json()["x_tenant"]["roles"] == [self.SUPER_ADMIN_ROLE]

    def test_post_request_with_correct_role_returns_201(self):
        payload = helpers.get_payload([self.SUPER_ADMIN_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 201
        assert response.json()["x_tenant"]["roles"] == [self.SUPER_ADMIN_ROLE]

    @classmethod
    def teardown_class(cls):
        flask_process.stop()



import os
import requests

from .. import flask_process, custom_token, helpers
from dotenv import load_dotenv


load_dotenv()


class TestScopeBasedPolicy:
    ENDPOINT = str(os.getenv("ENDPOINT"))
    REGULAR_USER_ROLE = "regular_user"

    @classmethod
    def setup_class(cls):
        flask_process.set_app_policies(
            [
                "token_based_policies.authlib_flask_oauth2_policy",
                "token_based_policies.tenant_token_roles_policy",
            ],
            ["scope_based_policy"],
        )
        flask_process.start()

    def setup_method(self):
        flask_process.assert_running()

    def test_request_with_insufficient_scopes_returns_403(self):
        payload = helpers.get_payload(["datateam"])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_regular_user_with_invalid_scopes_cannot_successfully_do_post_request(self):
        payload = helpers.get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.post(self.ENDPOINT, headers=headers)

        assert response.status_code == 403

    def test_request_with_sufficient_scopes_returns_200(self):
        payload = helpers.get_payload([self.REGULAR_USER_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        response_body = response.json()
        assert response.status_code == 200
        assert response_body["x_tenant"]["roles"] == [self.REGULAR_USER_ROLE]
        assert response_body["x_tenant"]["scopes"] == helpers.get_scopes(
            self.REGULAR_USER_ROLE
        )

    @classmethod
    def teardown_class(cls):
        flask_process.stop()



class AccessRestrictions:
    """
    A class containing properties used to restrict access.

    The value of the properties of this class are defined by authorization
    policies, and provides the ability to do attribute based access control
    (abac).

    Properties
    ----------
    filters : Unknown
        Object used to filter on attributes of documents.
    post_request_hook : (response: Unknown, *args: Unknown, **kwargs: Unknown) -> Unknown
        Function used to manipulate response data of a request.
    """

    def __init__(self):
        self._filters = None
        self._post_request_hook = lambda response, *args, **kwargs: response

    @property
    def filters(self):
        """Object used to filter on attributes of documents."""

        return self._filters

    @filters.setter
    def filters(self, filters):
        self._filters = filters

    @property
    def post_request_hook(self):
        """Function used to manipulate response data of a request."""

        return self._post_request_hook

    @post_request_hook.setter
    def post_request_hook(self, post_request_hook):
        self._post_request_hook = post_request_hook



import json
import os


def get_payload(roles):
    return {
        "azp": "inuits-policy-based-auth",
        "resource_access": {"inuits-policy-based-auth": {"roles": roles}},
    }


def get_scopes(role):
    with open(str(os.getenv("TEST_API_SCOPES")), "r") as scopes_file:
        return json.load(scopes_file)[role]



import os

from .. import flask_process
from .policy_loader import load_policies
from .global_policy_factory import (
    _policy_factory,
    apply_policies,
    authenticate,
    get_user_context,
)
from flask import Flask, make_response
from flask_restful import Api, Resource, request
from inuits_policy_based_auth import RequestContext
from logging import Logger


app = Flask(__name__)
api = Api(app)

load_policies(_policy_factory, Logger(""))


class GenericEndpoints(Resource):
    @apply_policies(RequestContext(request, ["read-entity", "update-entity"]))
    def get(self):
        user_context = get_user_context()
        (
            number_of_authenticate_calls,
            number_of_apply_policies_calls,
        ) = _get_number_of_decorator_calls()

        response_body = {
            "auth_objects": {
                "token": user_context.auth_objects.get("token"),
            },
            "email": user_context.email,
            "x_tenant": {
                "id": user_context.x_tenant.id,
                "roles": user_context.x_tenant.roles,
                "scopes": user_context.x_tenant.scopes,
                "raw": user_context.x_tenant.raw,
            },
            "bag": user_context.bag,
            "access_restrictions": {
                "filters": user_context.access_restrictions.filters
            },
            "number_of_authenticate_calls": number_of_authenticate_calls,
            "number_of_apply_policies_calls": number_of_apply_policies_calls,
            "request_context": {
                "generic_endpoint": {
                    "hash": hash(RequestContext(request)),
                    "serialized": str(RequestContext(request)._serialize()),
                }
            },
        }

        flask_process.clear_logs()
        return response_body

    @apply_policies(RequestContext(request))
    def post(self):
        user_context = get_user_context()
        (
            number_of_authenticate_calls,
            number_of_apply_policies_calls,
        ) = _get_number_of_decorator_calls()

        response_body = {
            "auth_objects": {
                "token": user_context.auth_objects.get("token"),
            },
            "email": user_context.email,
            "x_tenant": {
                "id": user_context.x_tenant.id,
                "roles": user_context.x_tenant.roles,
                "scopes": user_context.x_tenant.scopes,
                "raw": user_context.x_tenant.raw,
            },
            "bag": user_context.bag,
            "access_restrictions": {
                "filters": user_context.access_restrictions.filters
            },
            "number_of_authenticate_calls": number_of_authenticate_calls,
            "number_of_apply_policies_calls": number_of_apply_policies_calls,
            "request_context": {
                "generic_endpoint": {
                    "hash": hash(RequestContext(request)),
                    "serialized": str(RequestContext(request)._serialize()),
                }
            },
        }

        flask_process.clear_logs()
        return response_body

    @authenticate(RequestContext(request))
    def put(self):
        user_context = get_user_context()
        (
            number_of_authenticate_calls,
            number_of_apply_policies_calls,
        ) = _get_number_of_decorator_calls()

        response_body = {
            "auth_objects": {
                "token": user_context.auth_objects.get("token"),
            },
            "email": user_context.email,
            "x_tenant": {
                "id": user_context.x_tenant.id,
                "roles": user_context.x_tenant.roles,
                "scopes": user_context.x_tenant.scopes,
                "raw": user_context.x_tenant.raw,
            },
            "bag": user_context.bag,
            "access_restrictions": {
                "filters": user_context.access_restrictions.filters
            },
            "number_of_authenticate_calls": number_of_authenticate_calls,
            "number_of_apply_policies_calls": number_of_apply_policies_calls,
            "request_context": {
                "generic_endpoint": {
                    "hash": hash(RequestContext(request)),
                    "serialized": str(RequestContext(request)._serialize()),
                }
            },
        }

        flask_process.clear_logs()
        return response_body


class ConcreteEndpoints(GenericEndpoints):
    @apply_policies(RequestContext(request, ["read-entity", "update-entity"]))
    def get(self):
        response_body = super().get()
        response_body["request_context"]["concrete_endpoint"] = {
            "hash": hash(RequestContext(request, ["read-entity", "update-entity"])),
            "serialized": str(
                RequestContext(request, ["read-entity", "update-entity"])._serialize()
            ),
        }
        return make_response(response_body, 200)

    @apply_policies(RequestContext(request))
    def post(self):
        response_body = super().post()
        response_body["request_context"]["concrete_endpoint"] = {
            "hash": hash(RequestContext(request)),
            "serialized": str(RequestContext(request)._serialize()),
        }
        return make_response(response_body, 201)

    @authenticate(RequestContext(request))
    def put(self):
        response_body = super().put()
        response_body["request_context"]["concrete_endpoint"] = {
            "hash": hash(RequestContext(request)),
            "serialized": str(RequestContext(request)._serialize()),
        }
        return make_response(response_body, 200)


def _get_number_of_decorator_calls() -> tuple[int, int]:
    number_of_authenticate_calls = 0
    number_of_apply_policies_calls = 0

    with open(str(os.getenv("TEST_API_LOGS")), "r") as logs:
        for line in logs:
            line = line.strip()
            if line == "authenticate":
                number_of_authenticate_calls += 1
            elif line == "apply_policies":
                number_of_apply_policies_calls += 1

    return number_of_authenticate_calls, number_of_apply_policies_calls


api.add_resource(ConcreteEndpoints, "/")



import json

from inuits_policy_based_auth import BaseAuthenticationPolicy
from inuits_policy_based_auth.contexts import UserContext
from werkzeug.exceptions import Unauthorized


class TenantTokenRolesPolicy(BaseAuthenticationPolicy):
    """
    An authentication policy that defines the roles and scope from the token
    for the x-tenant.

    Parameters:
    -----------
    token_schema : dict
        Dict containing mappings between property <-> path.to.that.property.in.token.
    role_scope_mapping_filepath : str, optional
        Path to a JSON file containing a mapping of scopes to their corresponding roles.
    allow_anonymous_users : bool, optional
        A bool about whether anonymous users are allowed to do requests.
    """

    def __init__(
        self,
        token_schema: dict,
        role_scope_mapping_filepath=None,
        allow_anonymous_users=False,
    ):
        self._token_schema = token_schema
        self._role_scope_mapping = self.__load_role_scope_mapping(
            role_scope_mapping_filepath
        )
        self._allow_anonymous_users = allow_anonymous_users

    def authenticate(self, user_context: UserContext, _):
        """
        Obtain the user roles and scopes from the token and store
        them in the default x-tenant.

        Parameters:
        -----------
        user_context : UserContext
            The context of the user requesting authentication.
        request_context : RequestContext
            The context of the request.

        Returns:
        --------
        UserContext
            The user context with the authenticated user details added.

        Raises:
        -------
        Unauthorized
            If the authentication fails.
        """

        try:
            token = user_context.auth_objects["token"]
        except Exception:
            if self._allow_anonymous_users:
                return user_context
            else:
                raise Unauthorized("No token")

        flattened_token = user_context.flatten_auth_object(token)
        user_context.x_tenant.roles = flattened_token.get(
            self._token_schema["roles"], []
        )
        if not self._role_scope_mapping:
            return user_context
        for role in user_context.x_tenant.roles:
            try:
                user_context.x_tenant.scopes.extend(self._role_scope_mapping[role])
            except KeyError:
                continue
        return user_context

    def __load_role_scope_mapping(self, file):
        try:
            with open(file, "r") as role_scope_mapping:
                return json.load(role_scope_mapping)
        except Exception:
            pass



from abc import ABC, abstractmethod
from inuits_policy_based_auth.contexts import UserContext, RequestContext
from inuits_policy_based_auth.exceptions import (
    AuthenticateMethodDidNotReturnObjectOfTypeUserContextException,
)


class BaseAuthenticationPolicy(ABC):
    """
    An abstract class used as an interface for concrete implementations of
    authentication policies.

    Methods
    -------
    apply(user_context, request_context):
        Applies the policy.

    authenticate(user_context, request_context):
        Authenticates a user.
    """

    def apply(
        self, user_context: UserContext, request_context: RequestContext
    ) -> UserContext:
        """Applies the policy.

        Parameters
        ----------
        user_context : UserContext
            An object containing data about the authenticated user.
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        UserContext
            An object containing data about the authenticated user.

        Raises
        ------
        AuthenticateMethodDidNotReturnObjectOfTypeUserContextException:
            If the authenticate method does not return an object of type UserContext.
        Unauthorized:
            If the user is not authenticated.
        """

        user_context = self.authenticate(user_context, request_context)
        if not isinstance(user_context, UserContext):
            raise AuthenticateMethodDidNotReturnObjectOfTypeUserContextException()

        return user_context

    @abstractmethod
    def authenticate(
        self, user_context: UserContext, request_context: RequestContext
    ) -> UserContext:
        """Authenticates a user.

        Parameters
        ----------
        user_context : UserContext
            An object containing data about the user to be authenticated.
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        UserContext
            An object containing data about the authenticated user.

        Raises
        ------
        Unauthorized:
            If the user is not authenticated.
        """

        pass



class Tenant:
    """
    A class containing user data in the context of a specific tenant.

    Properties
    ----------
    id : str
        Id of the tenant the authenticated user is part of.
    roles : list[str]
        The roles of the authenticated user.
    scopes : list[str]
        The scopes of the authenticated user.
    raw : dict
        Raw object representation of the tenant the authenticated user is part of.
    """

    def __init__(self):
        self._id = ""
        self._roles = []
        self._scopes = []
        self._raw = {}

    @property
    def id(self):
        """Id of the tenant the authenticated user is part of."""

        return self._id

    @id.setter
    def id(self, id: str):
        self._id = id

    @property
    def roles(self):
        """The roles of the authenticated user."""

        return self._roles

    @roles.setter
    def roles(self, roles: list[str]):
        self._roles = roles

    @property
    def scopes(self):
        """The scopes of the authenticated user."""

        return self._scopes

    @scopes.setter
    def scopes(self, scopes: list[str]):
        self._scopes = scopes

    @property
    def raw(self):
        """
        Raw object representation of the tenant the authenticated user is part of.
        """

        return self._raw

    @raw.setter
    def raw(self, raw: dict):
        self._raw = raw



import inspect
import os
import re as regex

from functools import wraps
from inuits_policy_based_auth.authentication.base_authentication_policy import (
    BaseAuthenticationPolicy,
)
from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext
from inuits_policy_based_auth.exceptions import (
    PolicyFactoryException,
    NoUserContextException,
    NoAuthenticationPoliciesToApplyException,
    NoAuthorizationPoliciesToApplyException,
    InvalidFallbackKey,
    NoFallbackKeySet,
)
from werkzeug.exceptions import Unauthorized, Forbidden


class PolicyFactory:
    """
    A class used to apply policies.

    Parameters
    ----------
    user_context_setter_for_request_lifecycle : function
        A function provided by the API responsible for managing the user context
        throughout the HTTP request lifecycle. This responsibility is delegated to
        the API, rather than the library, to allow framework-specific implementations
        to handle more complex scenarios, such as those in multiprocess and
        multithreaded environments.

    Methods
    -------
    register_authentication_policy(key, policy)
        Appends a policy to the list of authentication policies to be applied
        mapped to a key.

    register_authorization_policy(key, policy)
        Appends a policy to the list of authorization policies to be applied
        mapped to a key.

    apply_policies(request_context)
        Applies registered policies to determine access.

    set_fallback_key_for_policy_mapping(key)
        Sets a fallback key for policy mapping if no key matches the first part
        of the module name of endpoints decorated by the policy factory.
    """

    def __init__(self, user_context_setter_for_request_lifecycle):
        self._authentication_policies: dict[str, list[BaseAuthenticationPolicy]] = {}
        self._authorization_policies: dict[str, list[BaseAuthorizationPolicy]] = {}
        self._fallback_key_for_policy_mapping = ""
        self._user_context_setter_for_request_lifecycle = (
            user_context_setter_for_request_lifecycle
        )

    def register_authentication_policy(
        self, key: str, policy: BaseAuthenticationPolicy
    ):
        """
        Appends a policy to the list of authentication policies to be applied mapped
        to a key.

        Parameters
        ----------
        key : str
            A key that matches the first part of the module name of endpoints decorated
            by the policy factory.
        policy : BaseAuthenticationPolicy
            A policy to be applied.
        """

        policies = self._authentication_policies.get(key)
        if policies:
            policies.append(policy)
        else:
            self._authentication_policies.update({key: [policy]})

    def register_authorization_policy(self, key: str, policy: BaseAuthorizationPolicy):
        """
        Appends a policy to the list of authorization policies to be applied mapped
        to a key.

        Parameters
        ----------
        key : str
            A key that matches the first part of the module name of endpoints decorated
            by the policy factory.
        policy : BaseAuthorizationPolicy
            A policy to be applied.
        """

        policies = self._authorization_policies.get(key)
        if policies:
            policies.append(policy)
        else:
            self._authorization_policies.update({key: [policy]})

    def set_fallback_key_for_policy_mapping(self, key: str):
        """
        Sets a key for policy mapping to fallback if no key matches the first part of
        the module name of endpoints decorated by the policy factory.

        Parameters
        ----------
        key : str
            A fallback key for policy mapping.

        Raises
        ------
        InvalidFallbackKey
            If provided key is not registered in either authentication policies,
            authorization policies, or both.
        """

        if (
            key != ""
            and self._authentication_policies.get(key)
            and self._authorization_policies.get(key)
        ):
            self._fallback_key_for_policy_mapping = key
        else:
            raise InvalidFallbackKey(key)

    def authenticate(self, request_context: RequestContext):
        """
        Applies registered authentication policies to determine access.

        Parameters
        ----------
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        function
            A decorator to decorate endpoints with.

        Raises
        ------
        NoAuthenticationPoliciesToApplyException
            If no authentication policies are registered.
        Unauthorized
            If user is not authenticated.
        NoFallbackKeySet
            If no fallback key for policy mapping is set.
        """

        def decorator(decorated_function):
            @wraps(decorated_function)
            def decorated_function_wrapper(*args, **kwargs):
                if len(self._authentication_policies) <= 0:
                    raise NoAuthenticationPoliciesToApplyException()

                if self.__is_test_environment(decorated_function):
                    import traceback
                    from flask import make_response

                    try:
                        self.__log_decorator("authenticate")
                        user_context = self._authenticate(
                            decorated_function, request_context
                        )
                    except Unauthorized as error:
                        self.__clear_logs()
                        raise error
                    except (TypeError, PolicyFactoryException) as error:
                        return make_response(
                            {
                                "message": str(error),
                                "stacktrace": traceback.format_exc(),
                            },
                            500,
                        )
                else:
                    user_context = self._authenticate(
                        decorated_function, request_context
                    )

                self._user_context_setter_for_request_lifecycle(user_context)
                return decorated_function(*args, **kwargs)

            return decorated_function_wrapper

        return decorator

    def apply_policies(self, request_context: RequestContext):
        """
        Applies registered policies to determine access.

        The first allowing policy will stop execution and allow access.
        The first denying policy will stop execution and deny access.
        If all policies are applied and none of them allowed or denied access
        (policy_context.access_verdict is None), then access is denied.

        Parameters
        ----------
        request_context : RequestContext
            An object containing data about the context of a request.

        Returns
        -------
        function
            A decorator to decorate endpoints with.

        Raises
        ------
        NoAuthenticationPoliciesToApplyException
            If no authentication policies are registered.
        NoAuthorizationPoliciesToApplyException
            If no authorization policies are registered.
        Unauthorized
            If user is not authenticated.
        Forbidden
            If user is not authorized.
        NoFallbackKeySet
            If no fallback key for policy mapping is set.
        NoUserContextException
            If there is no user auth data yet.
        """

        def decorator(decorated_function):
            @wraps(decorated_function)
            def decorated_function_wrapper(*args, **kwargs):
                if self.__is_test_environment(decorated_function):
                    import traceback
                    from flask import make_response

                    try:
                        self.__log_decorator("apply_policies")
                        user_context = self._apply_policies_decorated_function_wrapper_implementation(
                            decorated_function, request_context
                        )
                    except (Unauthorized, Forbidden) as error:
                        self.__clear_logs()
                        raise error
                    except (TypeError, PolicyFactoryException) as error:
                        return make_response(
                            {
                                "message": str(error),
                                "stacktrace": traceback.format_exc(),
                            },
                            500,
                        )
                else:
                    user_context = (
                        self._apply_policies_decorated_function_wrapper_implementation(
                            decorated_function, request_context
                        )
                    )

                self._user_context_setter_for_request_lifecycle(user_context)
                return decorated_function(*args, **kwargs)

            return decorated_function_wrapper

        return decorator

    def _apply_policies_decorated_function_wrapper_implementation(
        self, decorated_function, request_context: RequestContext
    ):
        if len(self._authentication_policies) <= 0:
            raise NoAuthenticationPoliciesToApplyException()
        if len(self._authorization_policies) <= 0:
            raise NoAuthorizationPoliciesToApplyException()

        user_context = self._authenticate(decorated_function, request_context)
        user_context = self._authorize(
            decorated_function, user_context, request_context
        )
        return user_context

    def _authenticate(self, decorated_function, request_context: RequestContext):
        user_context = UserContext()

        key = self._get_key_for_policy_mapping(
            self._authentication_policies, decorated_function
        )
        for policy in self._authentication_policies[key]:
            user_context = policy.apply(user_context, request_context)
            self._user_context_setter_for_request_lifecycle(user_context)

        return user_context

    def _authorize(
        self, decorated_function, user_context, request_context: RequestContext
    ):
        if not user_context:
            raise NoUserContextException()

        policy_context = PolicyContext()

        key = self._get_key_for_policy_mapping(
            self._authorization_policies, decorated_function
        )
        for policy in self._authorization_policies[key]:
            policy_context = policy.apply(policy_context, user_context, request_context)

            if policy_context.access_verdict:
                return user_context
            elif policy_context.access_verdict is False:
                break

        raise Forbidden()

    def _get_key_for_policy_mapping(self, policy_dict: dict, decorated_function) -> str:
        for key in policy_dict.keys():
            if regex.match(rf"^{key}.*", decorated_function.__module__) is not None:
                return key

        if self._fallback_key_for_policy_mapping:
            return self._fallback_key_for_policy_mapping

        raise NoFallbackKeySet()

    def __is_test_environment(self, decorated_function) -> bool:
        decorated_function_module = inspect.getmodule(decorated_function)
        if decorated_function_module:
            return (
                regex.match(
                    rf".*/inuits-policy-based-auth/{os.getenv('FLASK_APP')}$",
                    str(decorated_function_module.__file__),
                )
                is not None
            )

        return False

    def __log_decorator(self, name: str):
        with open(str(os.getenv("TEST_API_LOGS")), "a") as logs:
            logs.write(f"{name}\n")

    def __clear_logs(self):
        with open(str(os.getenv("TEST_API_LOGS")), "w") as logs:
            logs.write("")



import inspect


class ImmutableDict:
    """
    A dictionary-like object that cannot be modified after initialization.

    The keys and values of an ImmutableDict can be any hashable object.
    Values can also be dictionaries or lists containing hashable objects.

    This class can be used to create read-only dictionaries that cannot be
    accidentally modified, as well as for creating immutable configurations
    or objects that need to be serialized.
    """

    def __init__(self, data):
        """
        Initialize an ImmutableDict object with a dictionary or mapping object.

        Parameters
        ----------
        data : dict-like
            A dictionary or mapping object to initialize the ImmutableDict with.

        Returns
        -------
        ImmutableDict
            An ImmutableDict object with the initial data.

        Raises
        ------
        TypeError
            If the input data is not a dictionary or mapping object.
        """

        self._data = {}
        self._add_initial_data(data)

    def get(self, key):
        """
        Get the value associated with a key.

        Parameters
        ----------
        key : hashable
            The key to get the value for.

        Returns
        -------
        hashable or None
            The value associated with the key, or None if the key is not found.
        """
        return self._data.get(key)

    def add_key_value_pair(self, key, value):
        """
        Add a new key-value pair to the ImmutableDict.

        Parameters
        ----------
        key : hashable
            The key of the new key-value pair.
        value : hashable
            The value of the new key-value pair.

        Raises
        ------
        TypeError
            If the ImmutableDict is modified.
        """

        if key in self._data:
            raise TypeError("Cannot modify immutable dict")

        self._data[key] = value

    def __getitem__(self, key):
        """
        Get the value associated with a key.

        Parameters
        ----------
        key : hashable
            The key to get the value for.

        Returns
        -------
        hashable
            The value associated with the key.

        Raises
        ------
        KeyError
            If the key is not found in the ImmutableDict.
        """

        return self._data[key]

    def __setitem__(self, key, value):
        """
        Raise a TypeError if the ImmutableDict is modified.

        Raises
        ------
        TypeError
            If the ImmutableDict is modified.
        """

        if inspect.stack()[1][0].f_locals.get("self") is not self:
            raise TypeError("Cannot modify immutable dict")

        self._data[key] = value

    def __delitem__(self, _):
        """
        Raise a TypeError if the ImmutableDict is modified.

        Raises
        ------
        TypeError
            If the ImmutableDict is modified.
        """

        raise TypeError("Cannot modify immutable dict")

    def __repr__(self):
        """
        Get the string representation of the ImmutableDict.

        Returns
        -------
        str
            The string representation of the ImmutableDict.
        """

        return str(self._data)

    def _add_initial_data(self, data):
        """
        Recursively add the initial data to the ImmutableDict.

        Parameters
        ----------
        data : dict-like
            A dictionary or mapping object to add to the ImmutableDict.
        """

        try:
            for key, value in data.items():
                if isinstance(value, dict):
                    self._data[key] = ImmutableDict(value)
                elif isinstance(value, list):
                    self._data[key] = ImmutableDict._freeze_list(value)
                else:
                    self._data[key] = value
        except AttributeError:
            raise TypeError(f"{data} is not a dict-like object")

    @staticmethod
    def _freeze_list(list_to_be_frozen):
        """
        Recursively freeze a list of hashable objects.

        Parameters
        ----------
        list_to_be_frozen : list
            The list to be frozen.

        Returns
        -------
        tuple
            A tuple representing the frozen list, with all nested lists and dictionaries
            recursively frozen as well.
        """
        frozen_list = []

        for item in list_to_be_frozen:
            if isinstance(item, dict):
                frozen_list.append(ImmutableDict(item))
            elif isinstance(item, list):
                frozen_list.append(ImmutableDict._freeze_list(item))
            else:
                frozen_list.append(item)

        return tuple(frozen_list)



"""
Module for securing API endpoints based on policies.

This module provides a generic and highly customizable way of securing
API endpoints based on policies using a simple decorator.
When dynamically loading policies from a custom configuration file, this
module will allow you to add new policies and override existing ones
to make it exactly fit your needs.

Classes
-------
    BaseAuthenticationPolicy
        used as an interface for concrete implementations of authentication policies
    BaseAuthorizationPolicy
        used as an interface for concrete implementations of authorization policies
    RequestContext
        contains data about the context of a request
    PolicyFactory
        the main class, used to apply policies
"""

from inuits_policy_based_auth.authentication.base_authentication_policy import (
    BaseAuthenticationPolicy,
)
from inuits_policy_based_auth.authorization.base_authorization_policy import (
    BaseAuthorizationPolicy,
)
from inuits_policy_based_auth.contexts import RequestContext
from inuits_policy_based_auth.policy_factory import PolicyFactory



import os
import requests

from .. import flask_process, custom_token
from dotenv import load_dotenv


load_dotenv()


class TestAuthlibFlaskOauth2Policy:
    ENDPOINT = str(os.getenv("ENDPOINT"))

    def test_call_without_token_returns_401_when_anonymous_users_not_allowed(self):
        self.__start_flask_process()
        response = requests.get(self.ENDPOINT)
        assert response.status_code == 401

    def test_call_with_invalid_token_returns_401_when_anonymous_users_not_allowed(self):
        self.__start_flask_process()
        headers = custom_token.get_authorization_header({})
        response = requests.get(self.ENDPOINT, headers=headers)
        assert response.status_code == 401

    def test_call_with_valid_token_returns_200_when_anonymous_users_not_allowed(self):
        self.__start_flask_process()
        payload = {"azp": "inuits-policy-based-auth"}
        headers = custom_token.get_authorization_header(payload)
        response = requests.get(self.ENDPOINT, headers=headers)
        assert response.status_code == 200

    def test_call_without_token_returns_200_when_anonymous_users_allowed(self):
        self.__start_flask_process(allow_anonymous_users=True)
        response = requests.get(self.ENDPOINT)
        assert response.status_code == 200

    def test_call_with_invalid_token_returns_401_when_anonymous_users_allowed(self):
        self.__start_flask_process(allow_anonymous_users=True)
        headers = custom_token.get_authorization_header({})
        response = requests.get(self.ENDPOINT, headers=headers)
        assert response.status_code == 401

    def test_call_with_valid_token_returns_200_when_anonymous_users_allowed(self):
        self.__start_flask_process(allow_anonymous_users=True)
        payload = {"azp": "inuits-policy-based-auth"}
        headers = custom_token.get_authorization_header(payload)
        response = requests.get(self.ENDPOINT, headers=headers)
        assert response.status_code == 200

    def __start_flask_process(self, allow_anonymous_users=False):
        os.environ["ALLOW_ANONYMOUS_USERS"] = (
            "true" if allow_anonymous_users else "false"
        )
        flask_process.set_app_policies(
            ["token_based_policies.authlib_flask_oauth2_policy"], ["open_data_policy"]
        )
        flask_process.start()
        flask_process.assert_running()

    def teardown_method(self):
        flask_process.stop()



from .dummies import FakeHttpRequest
from inuits_policy_based_auth.authorization.policies.open_data_policy import (
    OpenDataPolicy,
)
from inuits_policy_based_auth.contexts import PolicyContext, RequestContext, UserContext


class TestOpenDataPolicy:
    def setup_method(self):
        self.fake_http_request = FakeHttpRequest()
        self.open_data_policy = OpenDataPolicy()
        self.policy_context = PolicyContext()
        self.user_context = UserContext()
        self.request_context = RequestContext(self.fake_http_request)

    def test_authorize_allows_access(self):
        self.fake_http_request._request_type = "GET"
        policy_context = self.open_data_policy.authorize(
            self.policy_context, self.user_context, self.request_context
        )
        assert policy_context.access_verdict == True

    def test_authorize_does_not_determine_access(self):
        self.fake_http_request._request_type = "POST"
        policy_context = self.open_data_policy.authorize(
            self.policy_context, self.user_context, self.request_context
        )
        assert policy_context.access_verdict is None



from flask import g
from inuits_policy_based_auth import PolicyFactory, RequestContext


def apply_policies(request_context: RequestContext):
    global _policy_factory
    return _policy_factory.apply_policies(request_context)


def authenticate(request_context: RequestContext):
    global _policy_factory
    return _policy_factory.authenticate(request_context)


def get_user_context():
    return g.get("user_context")


def user_context_setter(user_context):
    g.user_context = user_context


_policy_factory = PolicyFactory(user_context_setter)



import os
import requests

from .. import flask_process, custom_token, helpers
from dotenv import load_dotenv


load_dotenv()


class TestTenantTokenRolesPolicy:
    ENDPOINT = str(os.getenv("ENDPOINT"))
    SUPER_ADMIN_ROLE = "super_admin"

    def test_call_without_token_returns_401_when_anonymous_users_not_allowed(self):
        self.__start_flask_process()
        response = requests.get(self.ENDPOINT)
        assert response.status_code == 401

    def test_call_without_token_returns_200_when_anonymous_users_allowed(self):
        self.__start_flask_process(allow_anonymous_users=True)
        response = requests.get(self.ENDPOINT)

        response_body = response.json()
        assert response.status_code == 200
        assert response_body["x_tenant"]["id"] == ""
        assert response_body["x_tenant"]["roles"] == []
        assert response_body["x_tenant"]["scopes"] == []
        assert response_body["x_tenant"]["raw"] == {}

    def test_policy_sets_correct_roles_and_scopes_when_anonymous_users_not_allowed(
        self,
    ):
        self.__start_flask_process(load_oauth2_policy=True)
        payload = helpers.get_payload([self.SUPER_ADMIN_ROLE])
        headers = custom_token.get_authorization_header(payload)

        response = requests.get(self.ENDPOINT, headers=headers)

        response_body = response.json()
        assert response.status_code == 200
        assert response_body["x_tenant"]["id"] == ""
        assert response_body["x_tenant"]["roles"] == [self.SUPER_ADMIN_ROLE]
        assert response_body["x_tenant"]["scopes"] == helpers.get_scopes(
            self.SUPER_ADMIN_ROLE
        )
        assert response_body["x_tenant"]["raw"] == {}

    def __start_flask_process(
        self, allow_anonymous_users=False, load_oauth2_policy=False
    ):
        os.environ["ALLOW_ANONYMOUS_USERS"] = (
            "true" if allow_anonymous_users else "false"
        )
        authentication_policies = [
            "token_based_policies.tenant_token_roles_policy",
        ]
        if load_oauth2_policy:
            authentication_policies.insert(
                0, "token_based_policies.authlib_flask_oauth2_policy"
            )

        flask_process.set_app_policies(authentication_policies, ["open_data_policy"])
        flask_process.start()
        flask_process.assert_running()

    def teardown_method(self):
        flask_process.stop()



import json
import os

from importlib import import_module
from inuits_policy_based_auth import PolicyFactory
from inuits_policy_based_auth.exceptions import (
    PolicyFactoryException,
)
from logging import Logger


def load_policies(policy_factory: PolicyFactory, logger: Logger):
    apps = {}

    with open(str(os.getenv("TEST_API_CONFIGURATION")), "r") as configuration_file:
        apps = json.load(configuration_file)

    for app in apps:
        try:
            auth_type = "authentication"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy = __instantiate_authentication_policy(
                    policy_module_name, policy, logger
                )
                policy_factory.register_authentication_policy(
                    f"tests.integration.{app}", policy
                )

            auth_type = "authorization"
            for policy_module_name in apps[app]["policies"].get(auth_type):
                policy = __get_class(app, auth_type, policy_module_name)
                policy_factory.register_authorization_policy(
                    f"tests.integration.{app}", policy()
                )
        except Exception as error:
            raise PolicyFactoryException(
                f"Policy factory was not configured correctly: {str(error)}"
            ).with_traceback(error.__traceback__)

    policy_factory.set_fallback_key_for_policy_mapping("tests.integration.test_api")


def __get_class(app, auth_type, policy_module_name):
    locations = [
        policy_module_name,
        f"apps.{app}.policies.{auth_type}.{policy_module_name}",
        f"inuits_policy_based_auth.{auth_type}.policies.{policy_module_name}",
    ]
    for location in locations:
        try:
            module = import_module(location)
            break
        except ModuleNotFoundError:
            pass
    else:
        raise ModuleNotFoundError(f"Policy {policy_module_name} not found")

    policy_class_name = module.__name__.split(".")[-1].title().replace("_", "")
    policy = getattr(module, policy_class_name)
    return policy


def __instantiate_authentication_policy(policy_module_name, policy, logger: Logger):
    token_schema = __load_token_schema()

    if policy_module_name == "token_based_policies.authlib_flask_oauth2_policy":
        allow_anonymous_users = (
            True
            if os.getenv("ALLOW_ANONYMOUS_USERS", "false").lower() == "true"
            else False
        )
        return policy(
            logger,
            token_schema,
            os.getenv("STATIC_ISSUER"),
            os.getenv("STATIC_PUBLIC_KEY"),
            os.getenv("ALLOWED_ISSUERS"),
            allow_anonymous_users,
        )
    if policy_module_name == "token_based_policies.tenant_token_roles_policy":
        return policy(
            token_schema,
            os.getenv("ROLE_SCOPE_MAPPING", os.getenv("TEST_API_SCOPES")),
            (
                True
                if os.getenv("ALLOW_ANONYMOUS_USERS", "false").lower() == "true"
                else False
            ),
        )

    return policy()


def __load_token_schema() -> dict:
    token_schema_path = os.getenv(
        "TEST_API_TOKEN_SCHEMA", "src/tests/integration/test_api/token_schema.json"
    )
    with open(token_schema_path, "r") as token_schema:
        return json.load(token_schema)



from inuits_policy_based_auth.contexts.policy_context import PolicyContext
from inuits_policy_based_auth.contexts.request_context import RequestContext
from inuits_policy_based_auth.contexts.user_context import UserContext
